"""
Created on Friday 28.02 2018
Author(s): Kasper Emil Thorvaldsen


Modified in Spring 2020
By Jon Hvideberg Holte 

Main changes in Spring 2020:
- The script has been changed to only use Excel for both input files and output files.
  These new/modified functions include:
      - Store_SDP
      - Store_SIM
      - read_input_data
      - read_SV_data
      - read_pricemod_data
      
- The optimisation model has been changed to include wind power. This means that new variable,
  parameters and consstraints has been added. Also, modification in the existing elements are also done.


This script contains multiple functions used in the "main_initialize.py" script.

The following functions exist, and a short explanation of each:

----- opt_model ------
Creates an abstract model for a battery selling and buying energy to the grid.
It also sells primary reserve capacity if there is sufficient resources available.
There exists multiple constraints and piecewise-linear options.
The piecewise-linear options must be created with data, and therefore exists...
during the "create_instance" setup in the SDP and SIM part of the model/script.

----- SDP_run ------
Performs a backward Stochastic Dynamic Programming approach, using the model created.
Based on number of days, SOC-states, prices and such, the model will find optimal...
Storage values (or WV for hydropower) for each state and scenario.
This is done with multiple iterations, untill convergence occurs (if it does!)

----- SIM_run ------
Performs an actual simulation of random prices for each day, for a user-defined number of weeks.
The price scenarios are random, to create a realistic-ish approach.
The values from the SDP_run are used here, and gives output results.

----- Store_SDP ------
Stores SV-data and DVal-data in two Excel-files. 
One Excel-file is important for further use in the model (code) while the Excel-file is easier to analyse.
Storage values and DVal (Offset in value function) are stored in the files.
This will occur for every iteration, so the user can analyze each iteration.
Also, a general .dat file is created, storing general information about the SDP-run.

----- Store_SIM ------
The output results from SIM_run is taken into this script and stored in both two Excel-files. 
One Excel-file is important for further use in the model (code) while the Excel-file is easier to analyse. 
Relevant information that the user might want will be stored. For each timestep, for each day and for each week.
Also, a general .dat file is created, storing general information about the SIM-run.

----- read_input_data ------
This file will take in an Excel-file, and extract the information it contains.
These files contain information about the battery, the SDP/SIM setup, etc.
The data will be extracted and altered to be useable in the SDP/SIM run.

----- read_SV_data ------
This file will take in an Excel-file having the SV_data from a SDP-run, and extract...
both the SV-values and the DVal-values. This enables to perform the SIM run optimally.

----- read_price_and_wind_data ------
This file takes in an Excel-file containing wind and price data, and extract the data about the different price scenarios and wind scenarios.
It also extract scenario probabilities, which is used to find the weighted SV-values...
and for the SIM run to determine actual price scenarios.

----- print_gen_data ------
This file will print general data about the current setup. It informs the user of the battery,...
market and other information, so that the user is ensured that nothing is fishy.

"""

import pyomo.environ as pyo                                 #Import Pyomo
from pyomo.opt import SolverFactory                         #Import solvers
import numpy as np                                          #Import numpy for complex array handling
import random                                               #Import random for random elements
import time                                                 #Import to measure time usage
import sys                                                  #Used to allow exiting program (only used for testing)
from datetime import datetime                               #Used to find current time and date
import math                                                 #Used to round off numbers
import matplotlib.pyplot as plt                             #Used to allow plotting
import pandas as pd


def opt_model(inp_dict):              
    
    """
    Description of model:
    It is considered a prototypem adjustments might be implemented in the future
    
    Assumptions done:
    - All bilds are sold or bought
    - Neutral risk, price taker
    - Efficiency in the battery can be constant, or have a piecewise-linear behavior (user-defined)
    - Limits to the battery can also be interpreted as limits to the grid
    - Limits to the grid/battery can be constant for all timesteps, or be varying for each timestep
    - Reserve capacity sold is never utilized (currently at least)
    - Reserve capacity sold must be available in the battery (and be within limits)
    - When checking limits for reserve capacity, highest charging efficiency and lowest discharge efficiency are used. This avoids risks.
    - Storage values are piecewise-linear. Storage value for full battery is 0, but a point for 99.99 % exists that is != 0
    - Reserve capacity can be sold in blocks, in other words there can be a constraint for equal amount sold for multiple timesteps
    -
    
    Note that the piecewise linear functions are not declared in the abstract model, ...
    as it must be declared with data. This is done when the instance is created in both SDP and SIM.
    However, the model is declaring constraints based on if piecewise is used!
    
    """
    
    
    
    #Implement the abstract model
    

    model = pyo.AbstractModel()                             #Initialize abstract model


    """
    Set:
    As of right now, the following sets are considered to be included in the end:
    - TimeSteps [number of steps in the model]  -> TS
    """

    model.TS        = pyo.Set(ordered = True)               #Create timestep set

    #Parameters

    """
    Price data:
    Parameters consisting of economical data. Energy price, reservce
    capacity price and artificial price for power is included.
    Possible extensions might arise later.
    """

    model.Eprice    = pyo.Param(model.TS, mutable = True) 	#Create parameter for energy prices for each TS [NOK/MWh]

    model.Cprice    = pyo.Param(model.TS, mutable = True)	#Create parameter for reserve capacity prices for each TS [NOK/MWh]

    model.Aprice    = pyo.Param(mutable = True)             #Create artificial power price, considered constant [NOK/MWh]
    
    model.DVal      = pyo.Param(mutable = True)             #Offset in value function [NOK]
    
    model.T         = pyo.Param(model.TS, mutable = True)   #Length of each TimeStep TS [in hours]. Can be varying
    
    
    model.Shed_value    = pyo.Param(mutable = True)        #Create artificial price/value for shedding power so the model has an incentive for doing that if nessesary, considered constant [NOK/MW]

    
    """
    Battery Data:
    """

    model.BatCap        = pyo.Param(mutable = True)				#Capacity of the battery storage unit [MW]

    model.MDOD          = pyo.Param(mutable = True)				#Maximum depth of discharge. Limit lower level of the battery [p.u.]
    
    model.BatStart      = pyo.Param(mutable = True)				#Starting capacity of the battery at the beginning of the first TS [p.u.]

    model.ResCap        = pyo.Param(mutable = True)				#Reserve capacity sales limit [MW]
    
    model.ChaMax        = pyo.Param(mutable = True)				#Maximum charging output on the converter [MWh]

    model.DisMax        = pyo.Param(mutable = True)				#Maximum discharge output on the converter [MWh]
    
    model.Cha_in_max    = pyo.Param(mutable = True)             #Highest input charging value [MWh]
    
    model.Dis_in_max    = pyo.Param(mutable = True)             #Highes input discharge value [MWh]
    
    model.Eff_cha_max   = pyo.Param(mutable = True)             #Highest charging efficiency if using efficiency curve. If not using eff_curve, it is are equal to Eff defined in the file
    
    model.Eff_dis_min   = pyo.Param(mutable = True)             #Lowest discharging efficiency if using efficiency curve. If not using eff_curve, it is equal to Eff defined in the file
    
    """
    The following parameters are optional parameters limiting charge/discharge capacity for each timestep TS.
    If this is not necessary, a boolean input value can disable this.
    """
    
    
    model.ChaLim    = pyo.Param(model.TS,mutable = True)    #Maximum charging output on the converter for timestep TS [MWh]
    
    model.DisLim    = pyo.Param(model.TS,mutable = True)    #Maximum discharge output on the converter for timestep TS [MWh]
    
    model.Cha_in_Lim= pyo.Param(model.TS,mutable = True)    #Maximum charging input on the converter for timestep TS [MWh]
    
    """
    Piecewise linear parameters
    These are implemented in SDP_run and SIM_run if enabled. 
    """
    
    model.SOC_pts       = pyo.Param(mutable = True)             #1-Dimensional list containing piecewise linear points for SOC [values from 0-1]
    
    model.WV_pts        = pyo.Param(mutable = True)             #1-Dimensional list containing Storage Values for the corresponding SOC-state
    
    model.Power_pts     = pyo.Param(mutable = True)             #1-Dimensional list setting bought/discharged Power in the system [MWh]
    
    model.P_cha_pts     = pyo.Param(mutable = True)             #1-Dimensional list setting charged power based on bought and efficiency for given power
    
    model.P_dis_pts     = pyo.Param(mutable = True)             #1-Dimensional list setting discharged sold power based on discharged from battery and efficiency for given power


    """
    Parameters added because wind turbine included
    """
    model.P_transf_max  = pyo.Param(mutable = True)	            #Maximum power transfer to and from the grid [MWh]

    model.P_wind_rated  = pyo.Param(mutable = True)	            #The rated maximum power output to the wind turbine [MW]
    
    model.P_wind_output    = pyo.Param(model.TS, mutable = True)	#Maximum power output given in percent from the wind turbine fro timestep TS [p.u.]
   
    
    
    #Variables


    model.soc         = pyo.Var(model.TS, within = pyo.NonNegativeReals, bounds = (1-model.MDOD,1))       #SOC - State of Charge. States the remaining power in the battery at the end of TS[k] [p.u.]
    
    model.p_in_batt	  = pyo.Var(model.TS, bounds = (0,model.ChaMax))                                      #Variable deciding maximum flow stored into the battery in TS[k]. Only includes charging. [MWh]

    model.p_out_batt  = pyo.Var(model.TS, bounds = (0,model.Dis_in_max))                                  #Variable deciding maximum dischaged from the battery in TS[k]. Only includes discharging. [MWh]
    
    model.p_in_conv   = pyo.Var(model.TS, bounds = (0,model.Cha_in_max))                                  #Variable deciding maximum flow into the converter in TS[k]. Only includes charging. [MWh]

    model.p_out_conv  = pyo.Var(model.TS, bounds = (0,model.DisMax))                                      #Variable deciding maximum dischaged from the converter in TS[k]. Only includes discharging. [MWh]

    model.p_bought    = pyo.Var(model.TS, bounds = (0,model.P_transf_max))                                #Variable deciding maximum flow bought from the grid in TS[k].  [MWh]  
    
    model.p_sold      = pyo.Var(model.TS, bounds = (0,model.P_transf_max))                                #Variable deciding maximum sold power from the battery in TS[k].  [MWh]
    
    model.p_flow 	    = pyo.Var(model.TS, bounds = (-model.P_transf_max,model.P_transf_max))              #Power bought or sold from/to the battery in the current TS[k]. Can have positive and negative values (+ ->discharge) [MWh]

    model.cap         = pyo.Var(model.TS, within = pyo.NonNegativeReals, bounds = (0,model.ResCap))       #Total reserve capacity sold [MW]

    model.art_p		    = pyo.Var(model.TS, within = pyo.NonNegativeReals)                                  #Artificial power that is utilized when absolute necessary. Should not normally occur, used as a failsafe [MWh]

    model.end_value   = pyo.Var(within = pyo.NonNegativeReals)                                            #Value of remaining power in the battery, based on end of model timesteps and current Storage value [NOK]

    model.bat_end     = pyo.Var(bounds = ((1-model.MDOD),1))                                              #Variable used to set ending battery level [p.u]
    
    
    """
    Variables added because wind turbine included
    """
    
    model.p_wind          = pyo.Var(model.TS, within = pyo.NonNegativeReals)                            #Variable deciding the power output from the wind turbine in TS[k]. Includes both power stored and transfered [MWh]
    
    model.p_wind_store    = pyo.Var(model.TS, within = pyo.NonNegativeReals)                            #Variable deciding the power output from the wind turbine that is stored in storage unit in TS[k] [MWh]

    model.p_wind_transf   = pyo.Var(model.TS, within = pyo.NonNegativeReals)                            #Variable deciding the power output from the wind turbine that is transferred to grid in TS[k] [MWh]

    model.p_wind_shed    = pyo.Var(model.TS, within = pyo.NonNegativeReals)                            #Variable deciding the sheded power from the wind turbine TS[k] [MWh]  
    
    
    #Objective
    """
    Objective:
    Sum of:
    - Sold and bought energy
    - Sold reserve capacity
    - Bought artificial energy
    - Value of remaining energy
    - Offset in value function
    """


    def ObjRule(model):
        return( sum(model.Eprice[k]*model.p_sold[k]*model.T[k] - model.Eprice[k]*model.p_bought[k]*model.T[k]  + model.Cprice[k]*model.cap[k]*model.T[k]  - model.Aprice*model.art_p[k]*model.T[k] + model.Shed_value*model.p_wind_shed[k]*model.T[k] for k in model.TS)  + model.end_value + model.DVal)

    model.OBJ = pyo.Objective(rule = ObjRule, sense = pyo.maximize)


    #Constraints
        
    
    #For each timestep k, the ending storage value is equal to starting value and flow change
    def Storage_balance(model,i):    
        if i == model.TS.first(): #The storage level at the end of the first timestep is equal to current and change in stored power                        
            return(model.soc[model.TS.first()]*model.BatCap     - model.p_in_batt[model.TS.first()]*model.T[model.TS.first()]    + (model.p_out_batt[model.TS.first()])*model.T[model.TS.first()] - model.art_p[model.TS.first()]*model.T[model.TS.first()] == model.BatStart*model.BatCap)
        
        else:   #The storage level equality for the rest of the timesteps k
            return(model.soc[model.TS.prev(i)]*model.BatCap    + model.p_in_batt[i]*model.T[i]  - (model.p_out_batt[i])*model.T[i]   + model.art_p[i]*model.T[i] == model.soc[i]*model.BatCap)
    model.Storage_bal_const     = pyo.Constraint(model.TS, rule = Storage_balance)     #Declare constraint

    """ Modified """
    #Upper Capacity limit. Sets highest up-regulation of the battery based on remaining up-regulation in the converter (not storage)
    def Upper_cap_lim(model,i):
        if inp_dict["Hourly_cha/dis_limit"] == True:                        #The constraint will be based on a changing discharge limit on grid
            return(-(model.p_out_conv[i]*model.T[i]-model.p_in_conv[i]*model.T[i]) >= model.cap[i]*model.T[i] - model.DisLim[i]*model.T[i])
            
        else:                                                               #The constraint has a constant discharge limit on grid
            return(-(model.p_out_conv[i]*model.T[i]-model.p_in_conv[i]*model.T[i]) >= model.cap[i]*model.T[i] - model.DisMax*model.T[i]) 
    model.U_cap_lim_const = pyo.Constraint(model.TS,rule = Upper_cap_lim)   #Declare constraint
    
    """ Modified """
    #Sets lower capacity limit. Lowest down regulation converters, when enabling that the battery can go from discharge to charge!
    def Lower_cap_lim(model,i):
        if inp_dict["Hourly_cha/dis_limit"] == True:                        #The constraint will be based on a changing charge limit on grid
            return(-(model.p_out_conv[i]*model.T[i]-model.p_in_conv[i]*model.T[i]) <= -model.cap[i]*model.T[i] + model.Cha_in_Lim[i]*model.T[i])

        else:                                                               #The constraint has a constant charging limit on grid
            return(-(model.p_out_conv[i]*model.T[i]-model.p_in_conv[i]*model.T[i]) <= -model.cap[i]*model.T[i] + model.Cha_in_max*model.T[i])
    model.L_cap_lim_const = pyo.Constraint(model.TS, rule = Lower_cap_lim)  #Declare constraint

    #Sets upper capacity limit based on remaining power in the battery and the efficiency for charging
    def Storage_cap_u_lim(model,i):
        return(model.cap[i]*model.T[i] <= (model.BatCap - model.soc[i]*model.BatCap)/model.Eff_cha_max)
    model.Storage_cap_u_lim_const = pyo.Constraint(model.TS, rule = Storage_cap_u_lim)  #Declare constraint

    #Sets lower capacity limit based on remaining power in the battery and the efficiency for discharging
    def Storage_cap_l_lim(model,i):
        return(model.cap[i]*model.T[i] <= (model.soc[i]*model.BatCap-(1-model.MDOD)*model.BatCap)*model.Eff_dis_min)
    model.Storage_cap_l_lim_const = pyo.Constraint(model.TS, rule = Storage_cap_l_lim)  #Declare constraint

    #Connects p_flow to be discharged minus charged power for each step TS
    def Connect_P(model,i):
        return(model.p_flow[i]*model.T[i] == model.p_sold[i]*model.T[i] - model.p_bought[i]*model.T[i])
    model.Connect_P_const = pyo.Constraint(model.TS, rule = Connect_P)                  #Declare constraint
    
    #Sets the final battery level to be equal to the other variable declared to be this value. For usage in the piecewise linear function
    def BatEnd(model):
        return(model.bat_end == model.soc[model.TS.last()])
    model.batend_const = pyo.Constraint(rule=BatEnd)                                    #Declare constraint
    
    """ Modified """
    #Sets the rule for efficiency calculation when charging/discharging the battery
    if inp_dict["Efficiency_curve"] != True:    #If we do not want to have an efficiency curve based on discharge
        
        #Constraint for relation between power into the converter and stored power, using a constant efficiency
        def P_charge(model,i):
            return(model.p_in_batt[i]*model.T[i] == model.Eff_cha_max*model.p_in_conv[i]*model.T[i])
        model.P_charge_const = pyo.Constraint(model.TS,rule=P_charge)                   #Declare constraint
        
        #Constraint for relation between discharged power from the battery and the output power from converter, using a constant efficiency
        def P_discharge(model,i):
            return(model.p_out_conv[i]*model.T[i] == model.Eff_dis_min*model.p_out_batt[i]*model.T[i])
        model.P_discharge_const = pyo.Constraint(model.TS,rule=P_discharge)             #Declare constraint
    
    
    #Optional constraint including block sales of reserve capacity. 
    if inp_dict["Reserve_blocks"] == True:                              #If we want to have reserve capacity blocks
        def Cap_equality(model,i):  
            if i == model.TS.first():                                   #If first timestep, we skip a constraint
                return(pyo.Constraint.Skip)
            for n in inp_dict["Reserve_intervals"]:                     #If the current timestep is in the list, we skip it
                if n == i:
                    return(pyo.Constraint.Skip)                         
                    
            return(model.cap[i] == model.cap[model.TS.prev(i)])         #We declare the capacity for timestep k to be equal to k-1
            
        model.Cap_equality_const = pyo.Constraint(model.TS,rule = Cap_equality)         #Declare constraint
    
    
    
    """ New constraints when included wind """ 
    #Connects the power sold to the grid (p_sold) to be stored power from battery and power from wind turbine for each step TS
    def Connect_P_in(model,i):
        return(model.p_sold[i]*model.T[i] == model.p_out_conv[i]*model.T[i] + model.p_wind_transf[i]*model.T[i])
    model.Connect_P_in_const = pyo.Constraint(model.TS, rule = Connect_P_in)                  #Declare constraint

    #Connects the power into the converter (p_in_conv) to be power bought from grid and power from wind turbine for each step TS
    def Connect_P_out(model,i):
        return(model.p_in_conv[i]*model.T[i] == model.p_bought[i]*model.T[i] + model.p_wind_store[i]*model.T[i])
    model.Connect_P_out_const = pyo.Constraint(model.TS, rule = Connect_P_out)                  #Declare constraint

    #Connects p_wind to be the sum of power from wind turbine that are stored and transferred (sold) to grid for each step TS
    def Connect_P_wind(model,i):
        return(model.p_wind[i]*model.T[i] == model.p_wind_store[i]*model.T[i] + model.p_wind_transf[i]*model.T[i])
    model.Connect_P_wind_const = pyo.Constraint(model.TS, rule = Connect_P_wind)                  #Declare constraint
    
    #Limit the wind production (p_wind) to be the max power output from the wind turbine for each step TS    
    def P_wind_lim(model,i):
        if inp_dict["Throttle"] == True:                        #The constraint will be based on the possibility of throttle the wind turbine, so shedding variable is active (p_wind_shed) and thus possible to reduce the power produced 
            return(model.p_wind[i]*model.T[i] + model.p_wind_shed[i]*model.T[i] == model.P_wind_rated*model.P_wind_output[i]*model.T[i])
    
        else:                                                   #The not possible to throttle. The constraint sets the wind power to max output
            return(model.p_wind[i]*model.T[i] == model.P_wind_rated*model.P_wind_output[i]*model.T[i])
    model.P_wind_lim_const = pyo.Constraint(model.TS, rule = P_wind_lim)  #Declare constraint


    #Upper transfer limit for reserve capacity. Sets highest up-regulation of the battery based on remaining up-regulation in the transmission line
    def Upper_tranf_cap_lim(model,i):
        return(-model.p_flow[i]*model.T[i] >= model.cap[i]*model.T[i] - model.P_transf_max) 
    model.U_tansf_cap_lim_const = pyo.Constraint(model.TS,rule = Upper_tranf_cap_lim)   #Declare constraint
    
    
    #Sets lower transfer limit for reserve capacity. Sets lowest down-regulation of the battery based on availible down-regulation in the transmission line
    def Lower_tansf_cap_lim(model,i):
        return(-model.p_flow[i]*model.T[i] <= -model.cap[i]*model.T[i] + model.P_transf_max)
    model.L_tansf_cap_lim_const = pyo.Constraint(model.TS, rule = Lower_tansf_cap_lim)  #Declare constraint

    
    return(model)                   #We return the abstract model created
    
    
def SDP_run(DataDict,opt_mod):
    
    """
    Created on Friday 13.03 2018
    Author: Kasper Emil Thorvaldsen
    
    This function calculates optimal storage values (SV) for each possible scenario and SOC-state
    
    Assumptions done:
    - SOC-state is divided into discrete values
    
    
    """
    
    #Extract constant values from the dictionary
    MaxIter             = DataDict["MaxIter"]
    NR                  = DataDict["NR"]
    MaxDays             = DataDict["MaxDays"]
    MaxProb             = DataDict["MaxProb"]
    NR_list             = DataDict["NR_list"]
    
    #Fill in practical information to the dictionary to be used
    DataDict["SOC_pts"] = {None:NR_list}
    
    #Time storage
    Iter_start          = np.zeros((MaxIter))                       #Store start time for iteration
    Iter_end            = np.zeros((MaxIter))                       #Store time usage for iteration
    Day_start           = np.zeros((MaxIter,MaxDays))               #Store start time for day in iteration
    Day_end             = np.zeros((MaxIter,MaxDays))               #Store time usage for day in iteration
    
    #Empty lists of to be calculated values
    
    #List for storage values
    StoreV_list         = np.zeros((MaxIter,MaxDays+1,NR,MaxProb))  #Store the storage value
    
    #Lists for offset in value function
    DVal                = 0                                         #Store the offset value in battery
    DValTable           = np.zeros((MaxIter,MaxDays+1,MaxProb))     #Store the offset value in battery
    
    #Lists for results from the calculation
    Policydiff          = np.zeros((MaxIter))                       #Store the storage value deviation
    Objective_list      = np.zeros((MaxIter,MaxDays,NR,MaxProb))    #Store the objective results    
    Alphafrom           = np.zeros((MaxIter,MaxDays,NR,MaxProb))    #Store the weighted objective results

    

    
    #Set initial convergence check to be false
    Convergence = False
    
    for iter in range(0,MaxIter):                                           #For each iteration
        
        Iter_start[iter]      = time.time()                                 #Store starting time of iteration iter
        
        for day in range(MaxDays-1,-1,-1):                                  #For each day starting at the last day
            
            Day_start[iter,day]   = time.time()                             #Store starting time of day day
    
            for n in range(0,NR):                                           #For each segment of storage value
                
                
                DataDict["BatStart"] = {None:NR_list[n]}                    #Declare current battery level
                
                for pr in range(0,MaxProb):                                 #For each scenario
                    
                    print("Current computation: iter = " + str(iter+1)+ ", day = " + str(day+1) + ", n = " + str(n+1) + ", pr = "+ str(pr+1))
                    
                    #Create Storage value list for this scenario based on the different battery levels
                    StoreV_curr             = []
                    for i in range(0,len(NR_list)):
                        StoreV_curr.append(NR_list[i]*StoreV_list[iter,day+1,i,pr]*(DataDict["BatCap"][None]*DataDict["MDOD"][None]))
                    
                    #Set value for DVal
                    DVal = DValTable[iter,day+1,pr]
                    
                    #Declare data for the current scenario
                    DataDict["Eprice"]          = DataDict["Eprice_list"][day,pr]
                    DataDict["Cprice"]          = DataDict["Cprice_list"][day,pr]
                    DataDict["P_wind_output"]   = DataDict["Wind_prod"][day,pr]
                    DataDict["WV_pts"]          = {None:StoreV_curr}
                    DataDict["DVal"]            = {None:DVal}
                    
                    
                    """
                    Start computation of current scenario
                    """
                    
                    #Create instance
                    instance            = opt_mod.create_instance(data = {"":DataDict}, namespace = "") #Create instance, load data from dictionary
                    
                    #Load piecewise linear functions if user wants to use them
                    
                    #Piecewise linear function for value of stored power in battery
                    instance.End_level  = pyo.Piecewise(instance.end_value,instance.bat_end,pw_pts = instance.SOC_pts.value,pw_constr_type = 'EQ',f_rule = instance.WV_pts.value,pw_repn='SOS2',warning_tol=-1.2)
                    
                    """ Modified """
                    #Piecewise for non-constant efficiency, if true
                    if DataDict["Efficiency_curve"] == True:
                    
                            #Piecewise linear function of finding the stored power in the battery compared to a changing efficiency curve based on power into the converter
                            instance.PI = pyo.Piecewise(instance.TS,instance.p_in_conv,instance.p_in_batt,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_cha_pts.value,pw_repn='SOS2',warning_tol=-1.2)
                        
                            #Piecewise linear function of finding the output power from the converter compared to a changing efficiency curve based on discharged power from the battery
                            instance.PO = pyo.Piecewise(instance.TS,instance.p_out_batt,instance.p_out_conv,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_dis_pts.value,pw_repn='SOS2',warning_tol=-1.2)
  
                    #Set the solver
                    opt                 = SolverFactory('gurobi',solver_io="python")
					
                    #Give the solver a time limit
                    opt.options['timelimit'] = 20
					
                    #Run the optimization
                    results             = opt.solve(instance, load_solutions=True)
    
                    #Load results
                    #instance.solutions.load_from(results)
                    
                    #Store the objective value
                    Objective_list[iter,day,n,pr] = instance.OBJ()                    
                    
                #Calculate new storage values
                
                for pfrom in range(0,MaxProb):                  #For the current scenario position
                    Store_value = 0
                    for pto in range(0,MaxProb):                #Find the weighted storage value for all future scenario positions
                        Store_value = Store_value + DataDict["Prob"][pfrom,pto,day]*Objective_list[iter,day,n,pto]
                    Alphafrom[iter,day,n,pfrom] = Store_value   #Store the value
                    
                    if n == 0:                                  #If we are at the first storage segment, data stored as DVal
                        DValTable[iter,day,pfrom]       = Alphafrom[iter,day,n,pfrom]
                    
                    else:                                       #If not, calculate storage value for previous storage segment
                        StoreV_list[iter,day,n-1,pfrom] = (Alphafrom[iter,day,n,pfrom]-Alphafrom[iter,day,n-1,pfrom])/((NR_list[n]-NR_list[n-1])*DataDict["BatCap"][None])
        
            Day_end[iter,day] = time.time()-Day_start[iter,day] #Time usage for day
            print("Time Day: "+ str(Day_end[iter,day])+"\n")    #Print the time usage for user
            
        Iter_end[iter]  = time.time()-Iter_start[iter]          #Time usage for iteration
        print("Time Iteration: "+ str(Iter_end[iter])+"\n")     #Print time usage for iteration
        
        
        #If we are not at the first iteration, we compare this round with previous round
        if iter != 0:
            for day in range(0,MaxDays):
                for n in range(0,NR):
                    for pr in range(0,MaxProb):
                        #Find sum of deviation for all possible scenarios between iter and iter-1
                        Policydiff[iter] = Policydiff[iter] + abs(StoreV_list[iter,day,n,pfrom]-StoreV_list[iter-1,day,n,pfrom])
            
            print("Total deviation iteration "+str(iter+1) + ": "+ str(Policydiff[iter]))
                        
            #If the deviation is lowered than the user-specified value, we have gotten convergence
            if Policydiff[iter] <= DataDict["Deviation_SVal"]:
                Convergence = True
        
        #If we have not gotten convergence, we update the next iterations ending storage value
        if Convergence != True:
            if iter != (MaxIter-1): #To avoid storing in a non-existing list
                StoreV_list[iter+1,MaxDays,:,:] = StoreV_list[iter,0,:,:]
        
        #If we have gotten convergence, we exit the iteration-loop
        else:
            print("Convergence has occured in " + str(iter+1) + " iterations")
            
            break       #Get out of the iteration-loop
    
    #Round down data to decrease numbers on output
    for iter2 in range(0,DataDict["MaxIter"]):
        Iter_end[iter2] = round(Iter_end[iter2],5)
        for day in range(0,DataDict["MaxDays"]):
            Day_end[iter2,day] = round(Day_end[iter2,day],5)
            for n in range(0,DataDict["NR"]):
                for pr in range(0,DataDict["MaxProb"]):
                    StoreV_list[iter2,day,n,pr]  = round(StoreV_list[iter2,day,n,pr],5)
                    DValTable[iter2,day,pr]      = round(DValTable[iter2,day,pr],5)
                    
    
    #Create dictionary to return wanted data
    OutDict = {}
    OutDict["Iter_time"]    = Iter_end          #Store time usage
    OutDict["Day_time"]     = Day_end           #Store time usage
    OutDict["Policydiff"]   = Policydiff        #Store deviation
    OutDict["StoreV_list"]  = StoreV_list       #Store storage values
    OutDict["DValTable"]    = DValTable         #Store offset table 
    OutDict["NR_list"]      = NR_list           #Store storage segments
    OutDict["Convergence"]  = Convergence       #Store convergence results
    OutDict["End_iter"]     = iter              #Stores number of iterations done
    OutDict["Obj"]          = Objective_list    #Stores data about objective result
    
    return(OutDict)

    
def SIM_run(DataDict,opt_mod):
    """
    Created on Friday 06.04 2018
    Author: Kasper Emil Thorvaldsen
    
    
    """
    
    #Simplify some data for shorter lines in the code
    NumScen             = DataDict["NumScen"]
    MaxDays             = DataDict["MaxDays"]
    MaxProb             = DataDict["MaxProb"]
    MaxTS               = DataDict["MaxTS"]
    NR                  = DataDict["NR"]
    NR_list             = DataDict["NR_list"]
    
    DataDict["SOC_pts"] = {None:NR_list}                                #Store the segment list in dictionary
    
    
    print("Simulating the performance for " + str(NumScen) + " scenarios.")
    print("Random seed is " + str(DataDict["RandomSeed"]))              #To allow repetition of case
    
    #Create lists used in the calculation
    Scen_time           = np.zeros(NumScen)                             #List for time measuring
    Scen_start          = np.zeros(NumScen)                             #List for start time
    BatStart_list       = np.zeros((NumScen,MaxDays))                   #List for battery level at start of day and scen
    Prob_list           = np.zeros((NumScen,MaxDays))                   #List for probability occurances
    StoreV_list         = {}                                            #List for actual storage value lists to use
    DVal_list           = {}                                            #List for actual Dval-values to be used
    
    Prob_list[0,0]      = DataDict["Start_prob"]                        #Starting occurance, user-defined
    BatStart_list[0,0]  = DataDict["BatStart_SIM"]                      #Store user-defined battery start value
    
    OutDict             = {}                                            #Dictionary for output data
    
    #Multiple lists storing computed data during the process.
    soc_l               = np.zeros((NumScen,MaxDays,MaxTS))
    cap_l               = np.zeros((NumScen,MaxDays,MaxTS))
    p_in_batt_l         = np.zeros((NumScen,MaxDays,MaxTS))
    p_out_batt_l        = np.zeros((NumScen,MaxDays,MaxTS))
    p_in_conv_l         = np.zeros((NumScen,MaxDays,MaxTS))
    p_out_conv_l        = np.zeros((NumScen,MaxDays,MaxTS))
    p_bought_l          = np.zeros((NumScen,MaxDays,MaxTS))
    p_sold_l            = np.zeros((NumScen,MaxDays,MaxTS))
    p_flow_l            = np.zeros((NumScen,MaxDays,MaxTS))
    
    """ New list storing data after including wind turbine"""
    p_wind_l            = np.zeros((NumScen,MaxDays,MaxTS))
    p_wind_shed_l      = np.zeros((NumScen,MaxDays,MaxTS))
    p_wind_store_l      = np.zeros((NumScen,MaxDays,MaxTS))
    p_wind_transf_l     = np.zeros((NumScen,MaxDays,MaxTS))
    P_wind_max_l        = np.zeros((NumScen,MaxDays,MaxTS))
    
    art_p_l             = np.zeros((NumScen,MaxDays,MaxTS))
    ChaLim_l            = np.zeros((NumScen,MaxDays,MaxTS))
    DisLim_l            = np.zeros((NumScen,MaxDays,MaxTS))
    Eprice_l            = np.zeros((NumScen,MaxDays,MaxTS))
    Cprice_l            = np.zeros((NumScen,MaxDays,MaxTS))
    Eprofit_l           = np.zeros((NumScen,MaxDays,MaxTS))
    Cprofit_l           = np.zeros((NumScen,MaxDays,MaxTS))
    BatLevel            = np.zeros((NumScen,MaxDays,MaxTS))
    TimeLength          = np.zeros((NumScen,MaxDays,MaxTS))
    End_value           = np.zeros((NumScen))
    
    #Store end value for each day
    end_value_l         = np.zeros((NumScen,MaxDays))
    
    #Start creating the random choice of what probability will occur at each scenario and day
    #What it does: Creates a random variable between 0 and 1. Checks probability for
    #each possibility, and checks if that is greater than the random variable.
    #If it is, the current possibility is the one occuring, if now we move 1 more possibility
    #and accumulate the total probability.
    
    for s in range(0,NumScen):                                  #For each scenario
        for day in range(0,MaxDays):                            #For each day
        
            if (s == 0 and day == 0):                           #If first s and day
                prev_prob       = Prob_list[0,0]                #Store previous segment as current
            
            else:                                               #We are beyond first day and s
                if day == 0:                                    #If first day in a scenario
                    prev_prob   = Prob_list[s-1,MaxDays-1]      #Previous equal to last scenario
                
                else:                   
                    prev_prob   = Prob_list[s,day-1]            #Previous equal to previous day
                
                random_value    = random.random()               #Random value created
                cumulative_val  = 0                             #Total cumulative value
                
                for pr in range(0,MaxProb):                     #For each possible possibility
                    cumulative_val  = cumulative_val + DataDict["Prob"][prev_prob,pr,day]
                    now_prob        = pr                        #Current possibility
                    
                    if cumulative_val >= random_value:          #If probability higher than random value

                        break                                   #Break the for-loop, we have found the occurence
                
                Prob_list[s,day] = now_prob                     #Store occurence in a list
                
    #As the probabilities have been found, the storage value lists can be found now as well
    #Note, that this does not make the future known for the model. It is still uncertain!
    #It makes it somewhat easier to keep track of in the code.
    
    for s in range(0,NumScen):
        for day in range(0,MaxDays):
            StoreV_list[s,day]  = DataDict["StoreV_Sim"][day,Prob_list[s,day]]  #Store StoreV list
            DVal_list[s,day]    = DataDict["DVal_Sim"][day,Prob_list[s,day]]    #Store DVal value
    
    
    for s in range(0,NumScen):                                          #For each scenario
        
        print("Currently in scenario " + str(s+1))                      #Print current scenario
        Scen_start[s]   = time.time()                                   #Start time for scenario s
        
        for day in range(0,MaxDays):                                    #For each day in scenario s
            
            print("Scenario " + str(s+1) + ", day " + str(day+1))       #Print scenario and day
            
            StoreV_curr = []                                            #Create storage value list
            for i in range(0,len(NR_list)):
                StoreV_curr.append(NR_list[i]*StoreV_list[s,day][i]*(DataDict["BatCap"][None]*DataDict["MDOD"][None]))
                    
            
            #Declare values and lists for current day and possibility
            DataDict["BatStart"]        = {None:BatStart_list[s,day]}       #Declare starting battery level for s and day
            DataDict["Eprice"]          = DataDict["Eprice_list"][day,Prob_list[s,day]]
            DataDict["Cprice"]          = DataDict["Cprice_list"][day,Prob_list[s,day]]
            DataDict["P_wind_output"]      = DataDict["Wind_prod"][day,Prob_list[s,day]]

            DataDict["WV_pts"]      = {None:StoreV_curr}
            DataDict["DVal"]        = {None:DVal_list[s,day]}
            

            """
            Start computation of current scenario
            """
                    
            #Create instance
            instance            = opt_mod.create_instance(data = {"":DataDict}, namespace = "") #Create instance, load data from dictionary
                    
            #Load piecewise linear functions if user wants to use them
                    
            #Piecewise linear function for value of stored power in battery
            instance.End_level  = pyo.Piecewise(instance.end_value,instance.bat_end,pw_pts = instance.SOC_pts.value,pw_constr_type = 'EQ',f_rule = instance.WV_pts.value,pw_repn='SOS2',warning_tol=-1.2)
                    
            """ Modified """
            #Piecewise for non-constant efficiency, if true
            if DataDict["Efficiency_curve"] == True:
                
                #Piecewise linear function of finding the stored power compared to a changing efficiency curve based on bought power
                #instance.PI     = pyo.Piecewise(instance.TS,instance.p_in_batt,instance.p_bought,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_cha_pts.value,pw_repn = 'INC',warning_tol=-1.2)
                        
                #Piecewise linear function of finding the sold power compared to a changing efficiency curve based on discharged power
                #instance.PO     = pyo.Piecewise(instance.TS,instance.p_sold,instance.p_out_batt,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_dis_pts.value,pw_repn = 'INC',warning_tol=-1.2)


                #Piecewise linear function of finding the stored power in the battery compared to a changing efficiency curve based on power into the converter
                instance.PI = pyo.Piecewise(instance.TS,instance.p_in_conv,instance.p_in_batt,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_cha_pts.value,pw_repn='SOS2',warning_tol=-1.2)
                        
                #Piecewise linear function of finding the output power from the converter compared to a changing efficiency curve based on discharged power in the battery
                instance.PO = pyo.Piecewise(instance.TS,instance.p_out_batt,instance.p_out_conv,pw_pts = instance.Power_pts.value,pw_constr_type = 'EQ', f_rule = instance.P_dis_pts.value,pw_repn='SOS2',warning_tol=-1.2)
  
            #Set the solver
            opt                 = SolverFactory('gurobi',solver_io="python")
			
            #Give the solver a time limit
            opt.options['timelimit'] = 20
			
            #Run the optimization
            results             = opt.solve(instance, load_solutions=True)
            
            #Store next days starting battery level
            if day == MaxDays-1:
                End_value[s] = instance.end_value.value
                if s == NumScen-1:
                    a = 1337            #Placeholder, no usage for this value, to avoid error
                else:
                    BatStart_list[s+1,0]    = instance.bat_end.value    #If at end of day, value equal to start of next scenario    
            else:
                BatStart_list[s,day+1]      = instance.bat_end.value    #Ending value now equal to starting value next day.
                
            

            #Start storing computation data in the lists. Also rounded numbers
            j = 0       #Value to help store data for each timestep TS
            for i in instance.TS:                                               #For each timestep
                
                soc_l[s,day,j]      = round(instance.soc[i].value,5)            #Store storage level (end of timeste)
                
                BatLevel[s,day,j]   = round(soc_l[s,day,j]*DataDict["BatCap"][None],5)
                
                cap_l[s,day,j]      = round(instance.cap[i].value,5)            #Store capacity sold
                
                #Power flow data stored
                p_in_batt_l[s,day,j]  = round(instance.p_in_batt[i].value,5)        
                p_out_batt_l[s,day,j] = round(instance.p_out_batt[i].value,5)
                p_in_conv_l[s,day,j]  = round(instance.p_in_conv[i].value,5)        
                p_out_conv_l[s,day,j] = round(instance.p_out_conv[i].value,5)

                p_bought_l[s,day,j]   = round(instance.p_bought[i].value,5)
                p_sold_l[s,day,j]     = round(instance.p_sold[i].value,5)
                p_flow_l[s,day,j]     = round(instance.p_flow[i].value,5)
                
                art_p_l[s,day,j]      = round(instance.art_p[i].value,5)

                """ New list storing data after including wind turbine"""                
                p_wind_l[s,day,j]          = round(instance.p_wind[i].value,5)
                p_wind_shed_l[s,day,j]    = round(instance.p_wind_shed[i].value,5)

                p_wind_store_l[s,day,j]    = round(instance.p_wind_store[i].value,5)
                p_wind_transf_l[s,day,j]   = round(instance.p_wind_transf[i].value,5)
                P_wind_max_l[s,day,j]      = round(instance.P_wind_output[i].value*DataDict["P_wind_rated"][None],5)
                    
                #Line limit stored
                ChaLim_l[s,day,j]   = round(instance.ChaLim[i].value,5)
                DisLim_l[s,day,j]   = round(instance.DisLim[i].value,5)
                
                #Market prices stored
                Eprice_l[s,day,j]   = round(instance.Eprice[i].value,5)
                Cprice_l[s,day,j]   = round(instance.Cprice[i].value,5)
                
                #Possible profits stored
                Eprofit_l[s,day,j]  = round(instance.Eprice[i].value*instance.p_flow[i].value,5)
                Cprofit_l[s,day,j]  = round(instance.Cprice[i].value*instance.cap[i].value,5)
                
                TimeLength[s,day,j] = instance.T[i].value
                
                
                j = j+1            
            end_value_l[s,day]      = round(instance.end_value.value,5)           #Store ending value
            
            
            #End day

        
        Scen_time[s] = time.time()-Scen_start[s]                       #Store time usage
        print("Time usage scenario " + str(s+1) + ": " + str(Scen_time[s]))
        #End Scenario
        
    #Store lists in output dictionary
    OutDict["SOC"]          = soc_l
    OutDict["Cap"]          = cap_l
    OutDict["P_in_batt"]    = p_in_batt_l
    OutDict["P_out_batt"]   = p_out_batt_l
    OutDict["P_in_conv"]    = p_in_batt_l
    OutDict["P_out_conv"]   = p_out_batt_l
    OutDict["P_bought"]     = p_bought_l
    OutDict["P_sold"]       = p_sold_l
    OutDict["P_flow"]       = p_flow_l
    OutDict["ArtP"]         = art_p_l
    OutDict["ChaLim"]       = ChaLim_l
    OutDict["DisLim"]       = DisLim_l
    OutDict["Eprice"]       = Eprice_l
    OutDict["Cprice"]       = Cprice_l
    OutDict["Eprofit"]      = Eprofit_l
    OutDict["Cprofit"]      = Cprofit_l
    OutDict["end_value"]    = end_value_l       
    OutDict["Prob_list"]    = Prob_list
    OutDict["StoreV_l"]     = StoreV_list
    OutDict["Scen_time"]    = Scen_time
    OutDict["BatLevel"]     = BatLevel
    OutDict["TimeLength"]   = TimeLength
    OutDict["End_value"]    = End_value
    
    """ New list storing data after including wind turbine"""                
    OutDict["p_wind"]         = p_wind_l
    OutDict["p_wind_shed"]   = p_wind_shed_l

    OutDict["p_wind_store"]   = p_wind_store_l
    OutDict["p_wind_transf"]  = p_wind_transf_l
    OutDict["P_wind_max"]     = P_wind_max_l
    
    return(OutDict)         #Return the dictionary
    
    
def Store_SDP(DataDict):
    
    """
    Created in Spring 2020
    Author: Jon Hvideberg Holte
    Based on the code from 20.03.2018, written by: Kasper Emil Thorvaldsen     
    """
    
    print("Storing data from SDP to files")                     #Explaining command done
        
    """
    Start with storing general information in a .dat-file
    Information about time usage and iteration deviation
    This information does not need to be stored in Excel
    """
    
    #Store some tables for shorter name calling
    time_usage  = DataDict["Iter_time"]                         #Time usage for each iteration
    Policydiff  = DataDict["Policydiff"]                        #Storage value deviation between iterations
    StoreV      = DataDict["StoreV_list"]                       #Storage value lists
    DVal        = DataDict["DValTable"]                         #Offset in value function list
    NR          = np.array(DataDict["NR_list"])
    
    w_file = open("SDP_data.dat","w")                           #Create file to store general information
    
    created_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S') #Store current time and date
    
    #Pre-planned sentences, explaining the process that has been done
    w_file.write("This file contains general data regarding a SDP run.\n")
    w_file.write("The following can be said about the SDP run:\n")
    w_file.write("Computation done: " + created_time + "\n")
    w_file.write("Number of days: " + str(DataDict["MaxDays"]) + "\n")
    w_file.write("Number of battery storage levels: " + str(DataDict["NR_segments"]) + "\n")
    w_file.write("Number of scenarios: " + str(DataDict["MaxProb"]) + "\n")
    w_file.write("Number of TimeSteps: " + str(DataDict["MaxTS"]) + "\n")
    w_file.write("Convergence aquired: " + str(DataDict["Convergence"]) + "\n")
    
    #Write the time usage for each iteration that has been executed
    w_file.write("\nTime Usage for each iteration: \n")
    time_tot = 0
    for i in range(0,len(time_usage)):
        if time_usage[i] != 0:                                  #Only store information for actual performed iterations
            time_tot = time_tot + time_usage[i]
            w_file.write("Iteration " + str(i+1) + ": " + str(time_usage[i]) +"\n")
        
    w_file.write("Time usage total: " + str(time_tot) + "\n")
    
    #Write down the deviation in storage value for each iteration executed
    w_file.write("\nStorage value deviation for each iteration: \n")    
    for i in range(1,len(Policydiff)):
        if i <= DataDict["End_iter"]:                                  #Only store information for actual performed iterations
            w_file.write("Iteration " + str(i) + "-" + str(i+1) + ": " + str(Policydiff[i]) +"\n")
    
    w_file.close()                                              #Close the file as all data has been written into it
    
    
    """
    Now, storing SDP-generated data in an Excel-file
    These data will NOT be called upon for the SIM part of the setup
    One Excel file is for visual analysing and overview
    One Excel file is for data handling with a file that is easy to read into Python
    
    Firstly, the overview Excel-file:
    """
    
    #Decide what the name of the output file should be
    excel_file_o = "SV_data_overview.xlsx"
    
    # Create a Pandas Excel writer object using XlsxWriter as the engine. 
    writer = pd.ExcelWriter(excel_file_o, engine ='xlsxwriter')

    # Create xlsxwriter workbook object 
    workbook = writer.book 
            
    # Here we create several formats for headers and texts. 
    format_1 = workbook.add_format({'bold': True, 'italic' : True})
    format_2 = workbook.add_format({'bold': True,'font_color': 'red', 'align': 'center' ,'border': 1})

    # Writes the last iteration, and thus the optimal values    
    # Some info explaining the process in the begining of the sheet
    info = pd.DataFrame([[datetime.now().strftime('%Y-%m-%d %H:%M:%S')], ["days: " + str(DataDict["MaxDays"]) + " , probabilities: " + str(DataDict["MaxProb"])], ["days: " + str(DataDict["MaxDays"]) + " , probabilities: " + str(DataDict["MaxProb"])],(DataDict["NR_list"])], index=["Created on: ", "Parameters for StoreV: ", "Parameters for DVal: ", "The StoreV-points are for the SOC-points [p.u]: "])
    # Write the info-dataframe to the worksheet
    info.to_excel(writer, sheet_name ="SV_data", header = False)

    #For Storage values StoreV
    for day in range(0,DataDict["MaxDays"]):                #For each day
        for pr in range(0,DataDict["MaxProb"]):             #For each storage value segment
            df_SV = pd.DataFrame([StoreV[DataDict["End_iter"],day,:,pr]] , index=["Day: "+str(day+1)+" , NR* , Prob.: "+ str(pr+1)])                             #define the values of the current scenario in a dataframe
            df_SV.to_excel(writer, sheet_name ="SV_data",startrow=(6+pr)+(day*DataDict["MaxProb"]), header=False, index_label="Storage values")                 #write that dataframe to Excel
    
    #For offset in value function DVAL
    for day in range(0,DataDict["MaxDays"]):                #For each day
        for pr in range(0,DataDict["MaxProb"]):             #For each scenario
            df_DVal = pd.DataFrame([DVal[DataDict["End_iter"],day,pr]] , index=["Day: "+str(day+1)+" , Prob.: "+str(pr+1)] )                                        #define the values of the current scenario in a dataframe
            df_DVal.to_excel(writer, sheet_name ="SV_data",startrow=((DataDict["MaxDays"]*DataDict["MaxProb"]+8)+pr)+(day*DataDict["MaxProb"]), header=False)   #write that dataframe to Excel

    #Add some more info and labels to every worksheet and define format
    info_SV = ("Storage Values based on SOC, StoreV (SV) [NOK]")
    info_DVal = ("Offset in value function, DVal [NOK]")
    worksheet = writer.sheets["SV_data"]
    worksheet.write(5, 0, info_SV, format_2)
    worksheet.write((DataDict["MaxDays"]*DataDict["MaxProb"]+7), 0, info_DVal, format_2)
    worksheet.set_column('A:A', 41)         # Set the column width and format. 

    # In this for-loop we write the values for each iteration that has been executed
    for it in range(0,DataDict["End_iter"]+1):                  #For each iteration that has been executed
        # Some info explaining the process in the begining of every sheet
        info = pd.DataFrame([[datetime.now().strftime('%Y-%m-%d %H:%M:%S')], ["days: " + str(DataDict["MaxDays"]) + " , probabilities: " + str(DataDict["MaxProb"])], ["days: " + str(DataDict["MaxDays"]) + " , probabilities: " + str(DataDict["MaxProb"])],(DataDict["NR_list"])], index=["Created on: ", "Parameters for StoreV: ", "Parameters for DVal: ", "The StoreV-points are for the SOC-points [p.u]: "])
        # Write the info-dataframe to every worksheet
        info.to_excel(writer, sheet_name ="SV_data"+str(it+1),startrow=1, header = False)
        
        #For Storage values StoreV
        for day in range(0,DataDict["MaxDays"]):                #For each day
            for pr in range(0,DataDict["MaxProb"]):             #For each storage value segment
                df_SV = pd.DataFrame([StoreV[it,day,:,pr]] , index=["Day: "+str(day+1)+" , NR* , Prob.: "+ str(pr+1)] )           #define the current scenario in a dataframe
                df_SV.to_excel(writer, sheet_name ="SV_data"+str(it+1),startrow=(7+pr)+(day*DataDict["MaxProb"]), header=False, index_label="Storage values") #write that dataframe to Excel
        
        #For offset in value function DVAL
        for day in range(0,DataDict["MaxDays"]):                #For each day
            for pr in range(0,DataDict["MaxProb"]):             #For each scenario
                df_DVal = pd.DataFrame([DVal[it,day,pr]] , index=["Day: "+str(day+1)+" , Prob.: "+str(pr+1)] )                          #define the current scenario in a dataframe
                df_DVal.to_excel(writer, sheet_name ="SV_data"+str(it+1),startrow=((DataDict["MaxDays"]*DataDict["MaxProb"]+9)+pr)+(day*DataDict["MaxProb"]), header=False)   #write that dataframe to Excel
                   
        
        #Add some more info and labels to every worksheet and define format
        infotext = ("These values stored in this file is values created midway, and not necessarily the final values.")
        info_SV = ("Storage Values based on SOC, StoreV (SV) [NOK]")
        info_DVal = ("Offset in value function, DVal [NOK]")
        worksheet = writer.sheets["SV_data"+str(it+1)]
        worksheet.write(0, 0, infotext, format_1)
        worksheet.write(6, 0, info_SV, format_2)
        worksheet.write((DataDict["MaxDays"]*DataDict["MaxProb"]+8), 0, info_DVal, format_2)
        worksheet.set_column('A:A', 41)         # Set the column width and format. 
        
    # Close the Pandas Excel writer object and output the Excel file. 
    writer.save() 

    """
    The readable Excel file for further Python use:
    """
    #Decide what the name of the output file should be
    excel_file_r = "SV_data_read.xlsx"
    
    # Create a Pandas Excel writer object using XlsxWriter as the engine. 
    writer = pd.ExcelWriter(excel_file_r, engine ='xlsxwriter')

    # Create a array with ONLY the values for the last iteration
    SV = StoreV[DataDict["End_iter"]][0:DataDict["MaxDays"]]
    DV = DVal[DataDict["End_iter"]][0:DataDict["MaxDays"]]

    #For Storage values StoreV
    df_SV = pd.DataFrame(SV.reshape(-1, DataDict["MaxProb"], order='A'))      #define the values in a dataframe
    df_SV.to_excel(writer, sheet_name ="Storage values")                      #write that dataframe to Excel
        #Since the data in the array is a 3d-array, it must be reshaped into the dataframe
        #It is reshaped into rows=day(NR) and columns=probability
    
    #For offset in value function DVAL
    df_DVal = pd.DataFrame(DV.reshape(-1, DataDict["MaxProb"], order='A'))    #define the values  in a dataframe
    df_DVal.to_excel(writer, sheet_name ="DVal")                              #write that dataframe to Excel
        #Since the data in the array is a 3d-array, it must be reshaped into the dataframe
        #It is reshaped into rows=day and columns=probability

    # Close the Pandas Excel writer object and output the Excel file. 
    writer.save()

    print("General information is stored in the .dat-file: SDP_data.dat")
    print("The results are now stored in one Excel-file for overveiw: " + excel_file_o)
    print("The results are now stored in one Excel-file that are easy to read into Python: " + excel_file_r)


    return                                                      #Finished with the script
    
def Store_SIM(DataDict):
    """
    Created in Spring 2020
    Author: Jon Hvideberg Holte
    Based on the code from 09.04.2018, written by: Kasper Emil Thorvaldsen 
    """
    
    print("Storing SIM data to corresponding files")
    
    created_time    = datetime.now().strftime('%Y-%m-%d %H:%M:%S') #Store current time and date
    
    #Store some lists locally for shorter names
    time_usage      = DataDict["Scen_time"]    
    
    """
    Start with storing general information in a .dat-file
    This information does not need to be stored in Excel
    """
    
    w_file          = open("SIM_data.dat","w") 
    
    w_file.write("This file contains general data regarding a SIM run.\n")
    w_file.write("The following can be said about the SIM run:\n")
    w_file.write("Computation done: " + created_time + "\n")
    w_file.write("Number of days: " + str(DataDict["MaxDays"]) + "\n")
    w_file.write("Number of battery storage levels: " + str(DataDict["NR_segments"]) + "\n")
    w_file.write("Number of possibilities: " + str(DataDict["MaxProb"]) + "\n")
    w_file.write("Number of TimeSteps: " + str(DataDict["MaxTS"]) + "\n")
    w_file.write("Number of scenario run: " + str(DataDict["NumScen"]) + "\n")
    
    w_file.write("\nTime Usage for each scenario: \n")
    time_tot = 0
    for i in range(0,len(time_usage)):
        if time_usage[i] != 0:                                  #Only store information for actual performed iterations
            time_tot = time_tot + time_usage[i]
            w_file.write("Scenario " + str(i+1) + ": " + str(time_usage[i]) +"\n")
    
    w_file.write("Total time usage: " + str(time_tot) + "\n")
    w_file.close()
    
    
    """
    Store the data output from the simulation in two Excel files. 
    One in an orderly fashion for overview, the other file is ment to be read into Python for analysing purposes.
    First, the overveiw file:
    """
    #Decide what the name of the output file should be
    excel_file_1 = "SIM_overview.xlsx"
    
    # Create a Pandas Excel writer object using XlsxWriter as the engine. 
    writer = pd.ExcelWriter(excel_file_1, engine ='xlsxwriter')
    
    # Create xlsxwriter workbook object 
    workbook = writer.book 
            
    # Here we create several formats for headers and texts. 
    format_1 = workbook.add_format({'bold': True, 'italic' : True})
    format_2 = workbook.add_format({'bold': True,'font_color': 'red'})

    #Find the total profits for both Energy and Capacit
    EPW_day     = np.sum(DataDict["Eprofit"],axis = 2)      #Find daily profit
    EPW_week    = np.sum(EPW_day,axis = 1)                  #Find weekly profit

    CPW_day     = np.sum(DataDict["Cprofit"],axis = 2)      #Find daily profit
    CPW_week    = np.sum(CPW_day,axis = 1)                  #Find weekly profit
    
    EVAL = DataDict["End_value"]                            #Array of the value of the stored energy in the storage unit at the end of the week
    scenario_number = np.arange(1,DataDict["NumScen"]+1)    #Array of the number of scenarios

    #Store these values in a dataframe that can be written to Excel
    #Also include a dataframe that contains relevant data about the computation
    info = pd.DataFrame([[created_time], [DataDict["NumScen"]],[DataDict["MaxDays"]],[len(DataDict["NR_list"])],[DataDict["Eff_segments"]]], index=["Computation done:", "Number of Scenarios:", "Number of days:", "Number of SOC segments:","Number of efficiency segments:"])
    endVal = pd.DataFrame({'Scenrio nr.': scenario_number,'EPW': EPW_week, 'CPW': CPW_week, 'EVAL': EVAL}) 

    # Write the two main dataframes worksheets.
    info.to_excel(writer, sheet_name ='Info', startrow=2, header = False)
    endVal.to_excel(writer, sheet_name ='End values', startrow=3, index=False)
    
    #Add an information text at the top of the sheet for the main sheets
    infotext = ("General data about the computation")
    worksheet = writer.sheets['Info']
    worksheet.write(0, 0, infotext, format_1)
    worksheet.set_column('A:A', 30)         # Set the column width and format. 
    text1 = ("This sheet contains energy and capacity sales profit for each scenario, and the value of the stored energy at the end of the week for each scenario.")    
    text2 = ("EPW = Energy Profit weekly [NOK]. CPW = Capacity Profit weekly [NOK]. EVAL = the value of the stored energy in the storage unit at the end of the week [NOK].")
    worksheet = writer.sheets['End values']
    worksheet.write(0, 0, text1, format_1)
    worksheet.write(1, 0, text2)
    worksheet.set_column('A:A', 10)         #Set the column width and format. 

    # Storing data about performance in separate woorksheets
    for data in DataDict["list"]:                       #For each wished output parameter in the list
        for scen in range(0,DataDict["NumScen"]):           #For each scenario
            if data == "TimeStep":                              #This case requires a different storage setup
                df = pd.DataFrame.from_dict(DataDict["TS"][None])   #Store data in a dataframe and write it to excel
                df.columns = ["Scenario nr.: " + str(scen+1)]
                df.to_excel(writer, sheet_name = data, startcol=1+(scen*2), startrow=2, index=False) 
                
            else:
                df = pd.DataFrame.from_dict(DataDict[data][scen])   #Store data for each scenario in a dataframe and write it to excel
                df = df.rename(index = lambda x: x + 1, columns = lambda x: x + 1)  #increase all the row/coulmn index label by value 1 so they match days and time step 
                df.to_excel(writer, sheet_name = data, startrow=(2+scen*9), index_label="Days\Hours") 
                
                worksheet = writer.sheets[data]
                text = ("Scenario number: " + str(scen+1))          #Write each scenario number
                worksheet.write((1+scen*9),0, text, format_2)
                worksheet.set_column('A:A', 10)                     #Set the column width and format. 

        if data != "TimeStep":                  # Write a infotext for each sheet
            worksheet = writer.sheets[data]
            worksheet.write(0,0, "This sheet contains information about the " + data + " for the " + str(DataDict["NumScen"]) + " scenarios. Each sceario has " +str(DataDict["MaxDays"])+ " days and " +str(DataDict["MaxTS"])+ " time steps (hours).  ", format_1)
        
    
    # Close the Pandas Excel writer object and output the Excel file. 
    writer.save() 
    
    """ 
    Then, the Python read file: 
    """
    #Decide what the name of the output file should be
    excel_file_2 = "SIM_read.xlsx"
    
    # Create a Pandas Excel writer object using XlsxWriter as the engine. 
    writer = pd.ExcelWriter(excel_file_2, engine ='xlsxwriter')
    
    #Find the total profits for both Energy and Capacit
    #Since EPW_week and CPW_week are defined in the code above they are not defined here.

    EVAL = DataDict["End_value"]                            #Array of the value of the stored energy in the storage unit at the end of the week

    #Store these values in a dataframe that can be written to Excel
    endVal = pd.DataFrame({'EPW': EPW_week, 'CPW': CPW_week, 'EVAL': EVAL}) 

    # Write the end value dataframe into a worksheet.
    endVal.to_excel(writer, sheet_name ='End values')

    # Storing data about performance in separate woorksheets
    for data in DataDict["list"]:       #For each wished output parameter in the list
        if data != "TimeStep":              #The data in TimeStep is organised in another way
            df = pd.DataFrame(DataDict[data].reshape(-1, DataDict["MaxDays"]*DataDict["MaxTS"], order='A'))   #Store data for each parameter in a dataframe and write it to excel
            df.to_excel(writer, sheet_name = data)         
                #Since the data in the dictonary is a 3d-array, it must be reshaped into the dataframe
                #It is reshaped into rows=scenarios and columns=day*hour
            
    # Close the Pandas Excel writer object and output the Excel file. 
    writer.save() 
    
    print("General information about the simulation is stored in the .dat-file: SIM_data.dat")
    print("The results are now stored in a orderly way in the Excel-file: " + excel_file_1)
    print("The results are now stored in a Excel-file that are easy to read into Python: " + excel_file_2)
        
    return
    

def read_input_data(filename,DataDict):
    
    print("Reading data from file: " + filename)            #Inform user of what file is read
    """
    MODIFIED CODE - READS EXCEL
    
    Created in spring 2020 by Jon Hvideberg Holte
    
    Read the Excel-file with the setup data. 
    """
    sheets = pd.ExcelFile(filename).sheet_names             #Obtain the sheet names
    
    for sheet in sheets:
        #Handles the sheet with single parameter values. Could be both integer and float. This is must be specified in the Excel-file.
        if sheet == "Param" :                                       
            temp = pd.read_excel(filename,sheet_name = sheet)       #Import data from Excel-sheet to dataframe
            for col in temp.columns:                                #Iterating the columns in the sheet
                DataDict[col] = temp.at[0,col]                      #Storing the single parameter values from the dataframe in dictionary
                
        #Handles the sheet with a list of parameter values. Specified as float values.
        if sheet == "Param_list":
            temp = pd.read_excel(filename,sheet_name = sheet)       #Import data from Excel-sheet to dataframe
            for col in temp.columns:                                #Iterating the columns in the sheet
                DataDict[col] = temp[col].astype(float).tolist()    #Storing parameter list values from the dataframe and storing in dictionary
        
        #Handles the sheet with a list of parameter values that is Specified as strings (text).
        if sheet == "Param_list_str":
            temp = pd.read_excel(filename,sheet_name = sheet)       #Import data from Excel-sheet to dataframes
            for col in temp.columns:                                #Iterating the columns from the sheet
                temp_list = temp[col].dropna()                      #Removing blank space if the list in Excel are not the same length
                DataDict[col] = temp_list.astype(str).tolist()      #Converting the text from the dataframe into a list of strings and storing in dictionary

        #Handles the sheet with a sets of parameter values that is Specified as float values.    
        if sheet == "Set":
            temp = pd.read_excel(filename,sheet_name = sheet)  #Import data from Excel-files to dataframes
            #Store the data from the dataframe in a dictionary based on one set and then storing it in the main dictionary
            DataDict.update(pd.read_excel(filename,sheet_name = sheet, index_col=0).to_dict())
     
    return(DataDict)



def read_SV_data(filename):   
    """
    MODIFIED CODE - READS EXCEL
    
    Created in spring 2020 by Jon Hvideberg Holte
    
    Read the Excel-file with the SV data. This Excel-file is obtanied through the SDP-phase. 
    """
    print("Reading data from file: " + filename)    #Inform user of what file is read
    
    sheets = pd.ExcelFile(filename).sheet_names     #Obtain the sheet names
    
    for sheet in sheets:                #The loop goes through the sheets that are imported in the Excel-file
        if sheet == "Storage values":   #Search for the sheet with the SV data. Note that the sheet needs a spesific name!
            StoreV_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )   #Extract the data from Excel into a dataframe
        elif sheet == "DVal":           #Search for the sheet with the DVal data. Note that the sheet needs a spesific name!
            DVal_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )     #Extract the data from Excel into a dataframe        
        else:
            print("The SV file does not contain the required sheets! (Storage values) (DVal)")
            sys.exit() #Exit the script. Insufficient data found, so it's no use to continue. The Excel file don't contain the sufficient data.    
        
    MaxDays         = DVal_data.shape[0]                    #Maxdays is found from the DVal input
    MaxProb         = DVal_data.shape[1]                    #MaxProb is found from the DVal input
    NR              = int(StoreV_data.shape[0]/MaxDays)     #Number of NR are found from the StoreV
    day             = 0                                     #Set the value of days to 0 for correct iteration
    pr              = 0                                     #Set the value of days to 0 for correct iteration
    #Creates empty dictionaries
    DataDict = {}
    DVal = {}
    StoreV = {}
    
    #Iteration-loop for storing DVal in the correct way    
    for i in range (MaxDays):                   #Iterate through the days
        for j in range (MaxProb):               #Iterate through the total number of probabilities
            DVal[day,pr] = DVal_data[j][i]      #Store the values from the DVal-dataframe in the local dictionary, with correct day and prob
                            
            pr = pr+1                           #Increase the current pr-value by +1
                        
            if pr == MaxProb:                   #We check if the maximum prob has been reached
                pr = 0                          #If so, reset prob value
                day = day +1  
            
            DataDict["DVal"] = DVal             #Store the local dictionary in the main dictionary

    
    day             = 0     #Reset the value of days to 0 for correct iteration
    pr              = 0     #Reset the value of days to 0 for correct iteration
    
    #Iteration-loop for storing SV in the correct way                
    for i in range (MaxDays):                   #Iterate through the days
        for j in range (MaxProb):               #Iterate through the total number of probabilities
            StoreV[day,pr] = StoreV_data[j].tolist()         #Store the row from the SV-dataframe in the local dictionary
            StoreV[day,pr] = StoreV[day,pr][i*NR:(i+1)*NR]   #To obtain the correct storage values the lists has to be divided with the number of storage segments and then stored with correct day and prob  
                    
            pr = pr+1                           #Increase the current pr-value by +1
                        
            if pr == MaxProb:                   #We check if the maximum prob has been reached
                pr = 0                          #If so, reset prob value
                day = day +1
                
            DataDict["StoreV"] = StoreV         #Store the local dictionary in the main dictionary
    
    return(DataDict)    #Return the collected dictionary
    
    
    
    
def read_price_and_wind_data(filename,TS,DataDict):
    
    """
    MODIFIED CODE - READS EXCEL
    
    Created in spring 2020 by Jon Hvideberg Holte
    
    Read the Excel-file with the price-data and wind-data. 

    Note that the energyprice (Eprice), reserve capacity price (Cprice) and the wind power production (Wind_prod) is stored so that each combination is unique for each day.
    E.g.:
    day ? / prob 0:    E_low	C_low	W_low
    day ? / prob 1:    E_low	C_low	W_mid
    day ? / prob 2:    E_low	C_low	W_high
    day ? / prob 3:    E_low	C_mid	W_low
    day ? / prob 4:    E_low	C_mid	W_mid
    day ? / prob 5:    E_low	C_mid	W_high
    day ? / prob 6:    E_low	C_high	W_low
    day ? / prob 7:    E_low	C_high	W_mid
    day ? / prob 8:    E_low	C_high	W_high
    day ? / prob 9:    E_mid	C_low	W_low
    day ? / prob 10:   E_mid	C_low	W_mid
    day ? / prob 11:   E_mid	C_low	W_high
    ...
    day ? / prob 24:   E_high	C_high	W_low
    day ? / prob 25:   E_high	C_high	W_mid
    day ? / prob 26:   E_high	C_high	W_high
    """
    # Obtain the sheet names
    sheets = pd.ExcelFile(filename).sheet_names
    
    MaxDays         = DataDict["MaxDays"]           #Maxdays is found from dictionary
    MaxProb         = DataDict["MaxProb"]           #MaxProb is found from dictionary ( 3 for Eprice * 3 for price * 3 for Wind = 27 combinations)
    #NumWeeks        = int(math.sqrt(MaxProb))      #NumWeeks is found from the probability
    NumProb         = DataDict["NumProb"]           #The number of probabilities for one of the parameters (e.g. 3 for Eprice, 3 for price, 3 for Wind)
    
    for sheet in sheets:    #The loop goes through the sheets that are important in the Excel-file
        
        day         = 0     #Default value for current day is reset
        pr          = 0     #Default value for current prob is reset
    
        #Obtain the energy prices: 
        if sheet == "Eprice" :
            Eprice_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )   #Extract the data from Excel into a dataframe
            Eprice = {}     #Create a local dictionary
            
            for i in range (MaxDays):   #iterate through the days
                for j in range (i,MaxDays*NumProb,MaxDays):    #iterate through the total number of days with steps of days in weeks
                    for k in range(NumProb*NumProb):           #iterate through the number of combinations needed for this probability (3 for Cprice * 3 for Wind)
                        Eprice[day,pr] = dict(zip(TS, Eprice_data.iloc[j]))     #Store the row in dataframe in the dictionary, with correct day and prob
                            
                        pr = pr+1                               #Increase the current pr-value by +1
                        
                        if pr == MaxProb:                       #We check if the maximum prob has been reached
                            pr = 0                              #If so, reset prob value
                            day = day +1  
        
            #OBS! These iterations that extract information from the Excel sheet are designed so that the local dictionary should have the same shape that Kasper used    
            DataDict["Eprice"] = Eprice     #Store the local dictionary in the main dictionary
    
        #Obtain the reserve capacity prices:
        if sheet == "Cprice" :
            Cprice_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )
            Cprice = {}     #Create a local dictionary
            
            for i in range (MaxDays):       #iterate through the days
                for k in range(NumProb):       #iterate through the number of probabilites
                    for j in range (i,MaxDays*NumProb,MaxDays):      #iterate through the total number of days with steps of days in weeks  
                        for t in range(NumProb):       #iterate through the number of probabilities
                            Cprice[day,pr] = dict(zip(TS, Cprice_data.iloc[j]))     #Store the row in dataframe in the dictionary, with correct day and prob
                
                            pr = pr+1                               #Increase the current pr-value by +1
                        
                            if pr == MaxProb:                       #We check if the maximum prob has been reached
                                pr = 0                              #If so, reset prob value
                                day = day +1  
        
            #OBS! These iterations that extract information from the Excel sheet are designed so that the local dictionary should have the same shape that Kasper used
            DataDict["Cprice"] = Cprice     #Store the local dictionary in the main dictionary
            
        #Obtain the wind production:
        if sheet == "Wind" :
            Wind_prod_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )
            Wind_prod = {}     #Create a local dictionary
            
            for i in range (MaxDays):       #iterate through the days
                for k in range(NumProb*NumProb):       #iterate through the number of combinations needed for this probability (3 for Eprice * 3 for Cprice)
                    for j in range (i,MaxDays*NumProb,MaxDays):      #iterate through the total number of days with steps of days in weeks  
                        Wind_prod[day,pr] = dict(zip(TS, Wind_prod_data.iloc[j]))     #Store the row in dataframe in the dictionary, with correct day and prob
            
                        pr = pr+1                               #Increase the current pr-value by +1
                    
                        if pr == MaxProb:                       #We check if the maximum prob has been reached
                            pr = 0                              #If so, reset prob value
                            day = day +1  
        
            #OBS! These iterations that extract information from the Excel sheet are designed so that the local dictionary should have the same shape that Kasper used
            DataDict["Wind_prod"] = Wind_prod     #Store the local dictionary in the main dictionary

            
        
        #Obtain the probabilities:
        if sheet == "Probabilities" :
            Prob_data = pd.read_excel(filename,sheet_name = sheet, index_col=0 )
            Prob = {}   #Create a local dictionary
            
            for i in range (MaxProb*MaxProb):       #iterate through all the probabilities
                for j in range (MaxDays):       #iterate through the days
                    Prob[day,pr,j] = Prob_data.iloc[i,j]     #Store the row in dataframe in the dictionary, with correct day and prob
                pr = pr+1                               #Increase the current pr-value by +1
                
                if pr == MaxProb:                       #We check if the maximum prob has been reached
                    pr = 0                              #If so, reset prob value
                    day = day +1  
    
            #OBS! These iterations that extract information from the Excel sheet are designed so that the local dictionary should have the same shape that Kasper used
            DataDict["Prob"] = Prob     #Store the local dictionary in the main dictionary

        
    return(DataDict)                                    #Return the collected dictionary    
    
def print_gen_data(DataDict):
    
    
    """
    Created on: 26.04.2018
    Author: Kasper Emil Thorvaldsen
    """
    
    print("Showing general data of the current setup\n")
    
    #Data about wind turbine and grid connection capacity
    print("Wind Turbine and Grid Data:")
    
    print("Maximum transfer limti to/from the grid: " + str(DataDict["P_transf_max"][None]) + " MW")
    print("Wind turbine power rating: " + str(DataDict["P_wind_rated"][None]) + " MW")
    print("Allow the wind turbine to throttle power output? " + str(DataDict["Throttle"]))
    
    
    #Data about battery capacity
    print("\nBattery Data:")
    
    print("Battery Capacity: " + str(DataDict["BatCap"][None]) + " MWh")
    print("Maximum Depth of Discharge: " + str(DataDict["MDOD"][None]) + " of total capacity")
    print("Reserve Capacity limit: " + str(DataDict["ResCap"][None]) + " MW")
    print("Varying Charge/Discharge? " + str(DataDict["Hourly_cha/dis_limit"]))
   
    
    x           = list(range(0, 0+DataDict["MaxTS"]))
    
    
    #If hourly charge/discharge, the pattern is plotted. If not, a constant value is written
    if DataDict["Hourly_cha/dis_limit"]:
        cha_list    = []
        dis_list    = []
        
        
        for el in DataDict["TS"][None]:
            cha_list.append(DataDict["ChaLim"][el])
            dis_list.append(DataDict["DisLim"][el])
        
        plt.plot(x, cha_list, 'r')
        plt.plot(x, dis_list, 'b')
        plt.ylabel("Transfer limit [MW]")
        plt.xlabel("Rimestep")
        
        plt.legend(('Charge limit', 'Discharge limit'),
           loc='upper right')
        
        plt.show()
        
    else:
        print("Charge Limit: " + str(DataDict["ChaMax"][None]))
        print("Discharge Limit: " + str(DataDict["DisMax"][None]))
    
    #If varying efficiency, the pattern is plotted. If not, a constant value is written
    print("Varying Efficiency? " + str(DataDict["Efficiency_curve"]))
    if DataDict["Efficiency_curve"]:
        
        plt.plot(DataDict["Power_pts"][None], DataDict["P_cha_eff"], 'r')
        plt.plot(DataDict["Power_pts"][None], DataDict["P_dis_eff"], 'b')
        plt.ylabel("Efficiency [%]")
        plt.xlabel("Power output [MW]")
        
        plt.legend(('Charging efficiency', 'Discharging efficiency'),
           loc='lower right')
        
        plt.show()
    
    else:
        print("Constant battery efficiency: " + str(DataDict["Eff_cha_max"][None]))
    
    
    #If reserve blocks are permitted, it is written. If not, nothing is said
    if DataDict["Reserve_blocks"]:
        line = "Number of Reserve blocks: " + str(len(DataDict["Reserve_intervals"])+1)
        print(line)
        line = "Reserve blocks: "
        first_block = DataDict["TS"][None][0]
        second_block = ""
        for i in range(0,len(DataDict["Reserve_intervals"])):
            #line = line + DataDict["Reserve_intervals"][i] + " "
            
            for el in range(0,len(DataDict["TS"][None])):
                if DataDict["Reserve_intervals"][i] == DataDict["TS"][None][el]:
                    second_block = DataDict["TS"][None][el-1]
                    break
            line = line + first_block + "-" + second_block + ", "
            first_block = DataDict["Reserve_intervals"][i]
        
        second_block = DataDict["TS"][None][-1]
        line = line + first_block + "-" + second_block
        print(line)
    else:
        print("No reserve blocks enabled")
    
    print("\n")
    
    #Print some market data
    print("Market data:")
    print("Number of timesteps: " + str(DataDict["MaxTS"]))
    print("Number of days: " + str(DataDict["MaxDays"]))
    print("Number of probabilities per day: " + str(DataDict["MaxProb"]))
    
    print("Number of State-of-charge states: " + str(DataDict["NR"]))
    
    
    print("Length of each TimeStep [hours]: ")
    
    T_list    = []
    
    #Plot timestep length
    for el in DataDict["TS"][None]:
        T_list.append(DataDict["T"][el])
        
    plt.plot(x, T_list, 'r')
    plt.ylabel("Time span [hours]")
    plt.xlabel("Timestep")
       
    plt.show()

    
    print("\n")
    print("Number of maximum iterations for SDP: " + str(DataDict["MaxIter"]))
    print("Convergence criteria: " + str(DataDict["Deviation_SVal"]))
    
    print("\n")
    
    print("Number of weeks for SIM: " + str(DataDict["NumScen"]))
    print("Done with general overview")
    input("Press Enter to continue...")
    
    
    return