"""
Created on Friday 28.02 2018
Author: Kasper Emil Thorvaldsen

Modified spring 2020
By Jon Hvideberg Holte 

This script will create the setup necessary to perform the simulations.
The script will set up all data from the input-files, and ask the user to specify what to perform.
The menu will contain the following:
- Show general data on current setup
- Perform a Stochastic Dynamic Programming performance of the model (SDP), backwards (day x to day 1)
- Perform a simulation of the battery for a user-defined number of iterations (weeks), based on SDP-results

Please note that some of the input data from the files must be altered for optimal performance.

Main changes in Spring 2020:
- The script has been changed to only use Excel for both input files and output files.

"""

import  numpy as np
import  random
import  time
import  sys
import  pyomo.environ as pyo                                    #Import Pyomo
from    pyomo.opt import SolverFactory                          #Import solvers
import  os.path

import functions_model_w as functions_model                     #File containing functions created for this.



DataDict = {}                   #The master dictionary, where everything from A to Z is stored

#Declare name of files that we extract data from
DataDict["setup_file"] = "setup_w.xlsx"
DataDict["Price_and_wind_file"] = "price_wind_prob.xlsx"


#We declare name of the SV-values file
DataDict["SV_data_file"] = "SV_data_read.xlsx"


#If some of the input files should not exist, the script will find that and shut down the script
if  (os.path.exists(DataDict["setup_file"]) == True and
    os.path.exists(DataDict["Price_and_wind_file"]) == True):
    print("All files exist")

else:
    print("Some of the input files does not exist in the same folder, please find them!")
    sys.exit() #Exit the script. Insufficient data found, so it's no use to continue


#Extract data from the 2 Excel-files
functions_model.read_input_data(DataDict["setup_file"],DataDict)
functions_model.read_price_and_wind_data(DataDict["Price_and_wind_file"],DataDict["TS"],DataDict)

#Create the NR-segments based on user-input. MDOD is considered as well
NR_list = []

for i in range(0,DataDict["NR"]-2):
    NR_list.append((1-DataDict["MDOD"]) + i*(DataDict["MDOD"])/(DataDict["NR"]-2))
    NR_list[i] = round(NR_list[i],5)
NR_list.append(1-1/((DataDict["NR"]-2)*100))    #The 99.99 % SOC-state
NR_list.append(1)                               #The 100 % SOC-state
DataDict["NR_list"] = NR_list                   #Store the list


#Find the highest and lowest efficiency from efficiency curve. Will be altered later if efficiency curve is false
#The highest and lowest are used to always fulfill that there is enough capacity if reserve capacities were to be utilized.
#This simplification does affect certain other areas, but will then assure no unnecessary risks.

if DataDict["Efficiency_curve"]:
    DataDict["Eff_cha_max"] = {None:max(DataDict["P_cha_pts"].copy())}
    DataDict["Eff_dis_min"] = {None:min(DataDict["P_dis_pts"].copy())}
else:
    DataDict["Eff_cha_max"] = {None:DataDict["Eff"]}
    DataDict["Eff_dis_min"] = {None:DataDict["Eff"]}


#Check if hourly limits. Alter "chamax" and "dismax" to have higest value in the series.
if DataDict["Hourly_cha/dis_limit"]:
    key1 = max(DataDict["ChaLim"], key=DataDict["ChaLim"].get)
    key2 = max(DataDict["DisLim"], key=DataDict["DisLim"].get)
    DataDict["ChaMax"] = DataDict["ChaLim"][key1]
    DataDict["DisMax"] = DataDict["DisLim"][key2]
    
    
#The Limits for hourly or not hourly is based on converter output limit. So input must also consider efficiency losses (P_out = P_in/efficiency)
DataDict["Cha_in_Lim"] = {}
for i in DataDict["TS"]:
    if DataDict["Efficiency_curve"]:
        DataDict["Cha_in_Lim"][i] = DataDict["ChaLim"][i]/DataDict["Eff_cha_max"][None]
    else:
        DataDict["Cha_in_Lim"][i] = DataDict["ChaLim"][i]/DataDict["Eff"]


#Manipulate Efficiency data:

#Store efficiency points for general data illustration
DataDict["P_cha_eff"] = DataDict["P_cha_pts"].copy()
DataDict["P_dis_eff"] = DataDict["P_dis_pts"].copy()

if DataDict["Efficiency_curve"]:
    for i in range(0,len(DataDict["Power_pts"])):
        #DataDict["Power_pts"][i] = DataDict["Power_pts"][i]*DataDict["BatCap"]
        if DataDict["Power_pts"][i] == 0:
            DataDict["P_cha_pts"][i] = 0
            DataDict["P_dis_pts"][i] = 0

        else:
            DataDict["P_cha_pts"][i] = DataDict["Power_pts"][i]/DataDict["P_cha_pts"][i]
            DataDict["P_dis_pts"][i] = DataDict["Power_pts"][i]/DataDict["P_dis_pts"][i]

    
    DataDict["Cha_in_max"] =    {None:DataDict["P_cha_pts"][-1]}
    DataDict["Dis_in_max"] =    {None:DataDict["P_dis_pts"][-1]}

else:   #If not efficiency curve is enabled, the maximum inputs to the converter is defined here.
    DataDict["Cha_in_max"] =    {None:DataDict["ChaMax"]/DataDict["Eff"]}
    DataDict["Dis_in_max"] =    {None:DataDict["DisMax"]/DataDict["Eff"]}
    


print("Now creating Abstract Optimization Model")

#The Abstract model is generated and stored in the local name model
model = functions_model.opt_model(DataDict)

#Some data must be altered to be able to put into the model. They must be stored in a non-key dictionary
#This is usually dimesion-less parameters or lists
DataDict["TS"] = {None:DataDict["TS"]}
DataDict["BatCap"] = {None:DataDict["BatCap"]} 
DataDict["Aprice"] = {None:DataDict["Aprice"]}
DataDict["ResCap"] = {None:DataDict["ResCap"]}
DataDict["Eff"] = {None:DataDict["Eff"]}
DataDict["MDOD"] = {None:DataDict["MDOD"]}
DataDict["ChaMax"] = {None:DataDict["ChaMax"]}
DataDict["DisMax"] = {None:DataDict["DisMax"]}
#DataDict["BatStart_SIM"] = {None:DataDict["BatStart_SIM"]}
DataDict["Power_pts"] = {None:DataDict["Power_pts"]}
DataDict["P_dis_pts"] = {None:DataDict["P_dis_pts"]}
DataDict["P_cha_pts"] = {None:DataDict["P_cha_pts"]}

""" New parameters when including wind turbine """
DataDict["Shed_value"] = {None:DataDict["Shed_value"]}
DataDict["P_transf_max"] = {None:DataDict["P_transf_max"]}
DataDict["P_wind_rated"] = {None:DataDict["P_wind_rated"]}


#Change the name of Eprice and Cprice due to naming in the model for SDP and SIM
DataDict["Eprice_list"] = DataDict["Eprice"]
DataDict["Cprice_list"] = DataDict["Cprice"]
DataDict["T"] = DataDict["TimeLength"]          #The naming from the input file is Timelength, but the model uses T. Doing this to give user better understanding


print("Model created!")

#If the SV_data file should not exist, either it is named wrong or not made yet.
#If it isn't made, the SDP-run will be necesssary to perform before the SIM-run
if (os.path.exists(DataDict["SV_data_file"]) == True):
    print("SV_data found, option 2 not necessary unless wanted")

else:
    print("SV_data not found, option 2 will be necessary before option 3")
    print("The required file is: " + str(DataDict["SV_data_file"]))

        


ans = True
    
while ans:          #Create a menu for the options
    print ("""
    1. Print general data about current setup
    2. Run Strategy Phase (Stochastic Dynamic Programmering part)
    3. Run Simulation Phase
    4. Exit/Quit
    """)
    
    ans = input("What would you like to do? ") 
    if ans == "1": 
    
        print("--------------------------------------------------------------------------------")
        functions_model.print_gen_data(DataDict)
        
    elif ans == "2":
    
        print("--------------------------------------------------------------------------------")
        print("Running SDP model")
        
        #Run the SDP using the stored data
        SDP = functions_model.SDP_run(DataDict.copy(),model)
        
        #Declare some data to the dictionary SDP
        SDP["MaxDays"] = DataDict["MaxDays"]
        SDP["MaxProb"] = DataDict["MaxProb"]
        SDP["MaxTS"] = DataDict["MaxTS"]
        SDP["NR_segments"] = DataDict["NR"]
                
        #Store the output data in a SV_data-file
        functions_model.Store_SDP(SDP)
        
    elif ans == "3":
        
        print("--------------------------------------------------------------------------------")
        #Check if the SV_data file exist
        if (os.path.exists(DataDict["SV_data_file"]) != True):
            print("SV_data not found, either try again or use option 2 next time")
            sys.exit()
            
        #Declare random seed. Ensure that the same output occurs even with multiple SIM-runs
        random.seed(DataDict["RandomSeed"])
        
        #Extract SV_data 
        output = functions_model.read_SV_data(DataDict["SV_data_file"])
        
        #Store the data in the dictionary
        DataDict["StoreV_Sim"] = output["StoreV"]
        DataDict["DVal_Sim"] = output["DVal"]
        
        #Run the SIM using stored data
        SIM = functions_model.SIM_run(DataDict.copy(),model)
        
        
        #Store relevant data in the SIM-dictionary
        SIM["MaxDays"] = DataDict["MaxDays"]
        SIM["MaxProb"] = DataDict["MaxProb"]
        SIM["MaxTS"] = DataDict["MaxTS"]
        SIM["NumScen"] = DataDict["NumScen"]
        SIM["NR_segments"] = DataDict["NR"]
        SIM["NR_list"] = DataDict["NR_list"]
        SIM["Eff_segments"] = len(DataDict["P_cha_pts"])
        SIM["TS"] = DataDict["TS"]
        SIM["ChaMax"] = DataDict["ChaMax"]
        SIM["DisMax"] = DataDict["DisMax"]
        SIM["Hourly_cha/dis_limit"] = DataDict["Hourly_cha/dis_limit"]
        
        """ Modified """
        #Declare output list of what to store. Users can modify this if wanted.
        output_list = ["TimeStep", "TimeLength","BatLevel", "SOC", "ArtP","Eprice","Eprofit","P_flow","P_sold","P_bought","P_in_batt", "P_out_batt","Cprice","Cprofit","Cap","ChaLim","DisLim", "P_wind_max", "p_wind", "p_wind_shed" ,"p_wind_store", "p_wind_transf"]

        SIM["list"] = output_list

        #Store the data
        functions_model.Store_SIM(SIM)
        
    elif ans == "4":
        
        print("--------------------------------------------------------------------------------")
        print("\n Goodbye") 
        ans = False
    elif ans != "":
        
        print("--------------------------------------------------------------------------------")
        print("\n Not Valid Choice Try again") 
      
